from __future__ import annotations

import errno
import inspect
import json
import os
import pickle
import sys
import tempfile
from collections import defaultdict
from contextlib import ExitStack
from functools import wraps
from pathlib import Path
from stat import S_IREAD, S_IRGRP, S_IROTH
from typing import Any, Optional, Union, Tuple, Callable
from zipfile import ZipFile

import cv2 as cv
import numpy as np
import pandas as pd
from PIL import Image
from PIL.PngImagePlugin import PngInfo
from s3a.constants import REQD_TBL_FIELDS as RTF, PRJ_ENUMS, IO_TEMPLATES_DIR
from s3a.generalutils import augmentException, getCroppedImg, resize_pad, cvImsave_rgb, orderContourPts, classproperty, \
  cvImread_rgb, imgPathtoHtml, deprecateKwargs, DirectoryDict
from s3a.parameditors.table import TableData
from s3a.shims import typing_extensions
from s3a.structures import FilePath, GrayImg, \
  ComplexXYVertices, PrjParam, AnnParseError
from skimage import draw
from skimage.exposure import rescale_intensity
from utilitys import fns
from utilitys.fns import warnLater

from .helpers import serialize, deserialize

FilePathOrDf = Union[FilePath, pd.DataFrame]
# Values are strings
# noinspection PyTypeHints
_litLst = typing_extensions.Literal[PRJ_ENUMS.IO_IMPORT, PRJ_ENUMS.IO_EXPORT]


def _attrNameFmt(buildOrExport: _litLst, obj):
  def membership(el):
    for exclude in ['FileType', 'Opts', 'Types', 'Serialized']:
      if el.endswith(exclude):
        return False
    return el.startswith(buildOrExport)

  attrs = [attr.replace(buildOrExport, '') for attr in vars(obj) if membership(attr)]
  out = {}
  for attr in attrs:
    descrFmt = fns.pascalCaseToTitle(attr)
    fileFmt = descrFmt.replace(' ', '.').lower()
    out[fileFmt] = descrFmt + ' Files'
  return out


def _getImg_mapping_offset(inFileOrImg: Union[FilePath, GrayImg]):
  if isinstance(inFileOrImg, np.ndarray):
    return inFileOrImg, None, None
  try:
    image: Image.Image = Image.open(inFileOrImg)
    info = image.info
    image = np.array(image)
  except TypeError:
    # E.g. float image
    return cvImread_rgb(inFileOrImg, mode=cv.IMREAD_UNCHANGED), None, None
  # "Offset" present for numeric data, "mapping" present for textual data
  if 'mapping' in info:
    mapping = pd.Series(json.loads(info['mapping']), name=info.get('field', None))
    mapping.index = mapping.index.astype(int)
    return image, mapping, None
  if 'offset' in info:
    return image, None, int(info['offset'])
  # No extra metadata
  return image, None, None


def _writeImge_meta(
    outImg: np.ndarray,
    saveName: FilePath,
    mapping: pd.Series = None,
    offset: int = None,
):
  outImg = Image.fromarray(outImg)
  info = PngInfo()
  if mapping is not None:
    info.add_text('mapping', json.dumps(mapping.to_dict()))
    info.add_text('field', str(mapping.name))
  if offset is not None:
    info.add_text('offset', str(offset))
  outImg.save(saveName, pnginfo=info)


def _getPdExporters():
  members = inspect.getmembers(
      pd.DataFrame, lambda meth: inspect.isfunction(meth) and meth.__name__.startswith('to_')
  )
  return [mem[0].replace('to_', '') for mem in members]


def _getPdImporters():
  members = inspect.getmembers(
      pd.DataFrame, lambda meth: inspect.isfunction(meth) and meth.__name__.startswith('read_')
  )
  return [mem[0].replace('read_', '') for mem in members]


# Credit: https://stackoverflow.com/a/28238047/9463643
# class classorinstancemethod(classmethod):
#   def __get__(self, instance, type_):
#     descr_get = super().__get__ if instance is None else self.__func__.__get__
#     return descr_get(instance, type_)


class ComponentIO:
  """
  Exporter responsible for saving Component information to a file or object.
  Once created, users can extract different representations of components by
  calling exporter.exportCsv, exportPkl, etc. for those objects / files respectively.
  """
  @classproperty
  def roundTripTypes(self):
    """
    Not all IO types can export->import and remain the exact same dataframe afterwards.
    For instance, exporting a labeled image will discard all additional fields.
    This property holds export types which can give back the original dataframe after
    a round trip export->import.
    """
    # Since both import and export should have these keys, can use either dict
    return {k: self.importTypes[k] for k in ['csv', 'pkl']}

  tableData = TableData()
  """Table to use for import/export cross checks. This is how class and table field information is derived."""
  importOpts = {}
  """
  Propagated to every importByFileType call to provide user-specified defaults as desired
  """
  exportOpts = {}
  """
  Propagated to every exportByFileType call to provide user-specified defaults as desired
  """

  def __init__(self):
    # Propagate custom defaults to each desired function
    for typeDict, fnType in zip([self.importTypes, self.exportTypes],
                                [PRJ_ENUMS.IO_IMPORT, PRJ_ENUMS.IO_EXPORT]):
      for fileExt in typeDict:
        func = self._ioFnFromFileType(fileExt, fnType)
        setattr(self, func.__name__, self._ioWrapper(func))

  @classproperty
  def exportTypes(cls):
    """
    File types this class can export. {file type: descriptoin} useful for adding to
    file picker dialog
    """
    return _attrNameFmt(PRJ_ENUMS.IO_EXPORT, cls)

  @classproperty
  def importTypes(cls):
    """
    Types this class can import. {file type: descriptoin} useful for adding to
    file picker dialog
    """
    return _attrNameFmt(PRJ_ENUMS.IO_IMPORT, cls)

  def _getForeignTableData(self, ioFuncName: str):
    """Returns a TableData object responsible for importing / exporting data of this format"""
    out = TableData(IO_TEMPLATES_DIR / (ioFuncName.lower() + '.tblcfg'))
    # Make sure current factories from plugins still work
    for param, factory in self.tableData.factories.items():
      out.addFieldFactory(param, factory)
    return out

  @classmethod
  def ioFileFilter(
      cls,
      which=PRJ_ENUMS.IO_ROUND_TRIP,
      typeFilter='',
      allFilesOpt=True,
      **extraOpts,
  ):
    """
    Helper for creating a file filter out of the handled IO types. The returned list of
    strings is suitable for inserting into a QFileDialog.

    :param which: Whether to generate filters for build types, export types, or round trip
    :param typeFilter: type filter for handled io types. For instanece, if typ='png', then
      a file filter list with only 'id.png' and 'class.png' will appear.
    :param allFilesOpt: Whether to add an "All Files (*.*) option to the dropdown menu
    :param extraOpts; Extra file types to include in the filter
    """
    ioDict = {
        PRJ_ENUMS.IO_ROUND_TRIP: cls.roundTripTypes,
        PRJ_ENUMS.IO_IMPORT: cls.importTypes,
        PRJ_ENUMS.IO_EXPORT: cls.exportTypes
    }[which]
    if isinstance(typeFilter, str):
      typeFilter = [typeFilter]
    fileFilters = []
    for typ, info in dict(**ioDict, **extraOpts).items():
      if any([t in typ for t in typeFilter]):
        fileFilters.append(f'{info} (*.{typ})')
    if allFilesOpt:
      fileFilters.append('All Files (*.*)')
    return ';;'.join(fileFilters)

  def _ioWrapper(self, func: Callable):
    """Wraps build and export functions to provide defaults specified by build/exportOpts before the function call"""
    which = PRJ_ENUMS.IO_IMPORT if func.__name__.startswith(
        PRJ_ENUMS.IO_IMPORT
    ) else PRJ_ENUMS.IO_EXPORT

    @wraps(func)
    def wrapper(*args, **kwargs):
      useOpts = self.importOpts if which is PRJ_ENUMS.IO_IMPORT else self.exportOpts
      # Turn args into kwargs
      fnKwargs = {**useOpts, **kwargs}
      return func(*args, **fnKwargs)

    return wrapper

  def exportByFileType(
      self,
      compDf: pd.DataFrame,
      outFile: Union[str, Path],
      verifyIntegrity=True,
      **exportArgs,
  ):
    outFile = Path(outFile)
    outFn = self._ioFnFromFileType(outFile, PRJ_ENUMS.IO_EXPORT)

    ret = outFn(compDf, outFile, **exportArgs)
    if verifyIntegrity and outFile.suffix[1:] in self.roundTripTypes:
      matchingCols = np.setdiff1d(compDf.columns, [RTF.INST_ID, RTF.SRC_IMG_FILENAME])
      loadedDf = self.importByFileType(outFile)
      # For some reason, there are cases in which numy comparison spots false errors and cases where
      # pandas spots false errors. Only a problem if both see problms
      dfCmp = np.array_equal(loadedDf[matchingCols], compDf[matchingCols]) \
              or loadedDf[matchingCols].equals(compDf[matchingCols])
      problemCells = defaultdict(list)

      if not dfCmp:
        dfA = loadedDf[matchingCols]
        dfB = compDf[matchingCols]
        for ii in range(len(dfA)):
          for jj in range(len(dfA.columns)):
            if not np.array_equal(dfA.iat[ii, jj], dfB.iat[ii, jj]):
              problemCells[compDf.at[dfB.index[ii], RTF.INST_ID]].append(str(matchingCols[jj]))
        # The only way to prevent "truth value of array is ambiguous" is cell-by-cell iteration
        problemMsg = [f'{idx}: {cols}' for idx, cols in problemCells.items()]
        problemMsg = '\n'.join(problemMsg)
        # Try to fix the problem with an iloc write
        warnLater(
            '<b>Warning!</b> Saved components do not match current component'
            ' state. This can occur when pandas incorrectly caches some'
            ' table values. Problem cells (shown as [id]: [columns]):\n' + f'{problemMsg}\n'
            f'Please try manually altering these values before exporting again.', UserWarning
        )
    return ret

  def importByFileType(
      self, inFile: Union[str, Path], imShape: Tuple[int] = None, strColumns=False, **importArgs
  ):
    buildFn = self._ioFnFromFileType(inFile, PRJ_ENUMS.IO_IMPORT)
    outDf = buildFn(inFile, imShape=imShape, **importArgs)
    if strColumns:
      outDf.columns = list(map(str, outDf.columns))
    return outDf

  def _ioFnFromFileType(self,
                        fpath: Union[str, Path],
                        buildOrExport=_litLst,
                        missingOk=False) -> Optional[Callable]:
    fpath = Path(fpath)
    fname = fpath.name
    cmpTypes = np.array(list(_attrNameFmt(buildOrExport, type(self))))
    typIdx = [fname.endswith(typ) for typ in cmpTypes]
    if not any(typIdx):
      raise IOError(f'Not sure how to handle file {fpath.name}')
    fnNameSuffix = cmpTypes[typIdx][-1].title().replace('.', '')
    outFn: Callable = getattr(self, buildOrExport + fnNameSuffix, None)
    if outFn is None and not missingOk:
      raise ValueError(f'Full I/O specification missing for type {fnNameSuffix}')
    return outFn

  def _pandasSerialImport(self, inFileOrDf: Union[str, Path, pd.DataFrame], **pdImportArgs):
    if isinstance(inFileOrDf, pd.DataFrame):
      serialDf = inFileOrDf
    else:
      fType = Path(inFileOrDf).suffix.lower().replace('.', '')
      importFn = getattr(pd, f'read_{fType}', None)
      if importFn is None:
        raise ValueError(
            f'File type {fType} cannot be handled by the serial importer.'
            f' Must be one of {",".join(_getPdImporters())}'
        )
      # Special case: csv imports need to avoid interpreting nan results
      pdImportArgs.update(na_filter=False, dtype=str)
      acceptedArgs = inspect.signature(importFn).parameters
      useArgs = pdImportArgs.keys() & acceptedArgs
      serialDf = importFn(inFileOrDf, **{k: pdImportArgs[k] for k in useArgs})
    return serialDf

  def _pandasSerialExport(
      self,
      exportDf: pd.DataFrame,
      outFile: Union[str, Path] = None,
      readOnly=True,
      **pdExportArgs
  ):
    if outFile is None:
      return

    defaultExportParams = {
        'na_rep': 'NaN',
        'float_format': '{:0.10n}',
        'index': False,
    }
    outPath = Path(outFile)
    outPath.parent.mkdir(exist_ok=True, parents=True)
    exporter = outPath.suffix.lower().replace('.', '')

    defaultExportParams.update(pdExportArgs)
    exportFn = getattr(exportDf, f'to_{exporter}', None)
    if exportFn is None:
      raise ValueError(
          f'Exporter {exporter} not recognized. Acceptable options:\n' +
          ', '.join(_getPdExporters())
      )

    with np.printoptions(threshold=sys.maxsize):
      exportFn(outFile, index=False)
    if readOnly:
      outPath.chmod(S_IREAD | S_IRGRP | S_IROTH)

  # -----
  # Export options
  # -----
  def exportSerialized(
      self, compDf: pd.DataFrame, outFile: Union[str, Path] = None, readOnly=True, **pdExportArgs
  ):
    """
    Converts dataframe into a string-serialized version and uses pandas to write it to disk.

    :param compDf: Dataframe to export
    :param outFile: Name of the output file location. If *None*, no file is created. However,
      the export object (string dtype dataframe) will still be created and returned.
      The file suffix can be any option supported by a pandas exporter. This can be
      csv, json, feather, etc.
      Note: pickle is a special case. In some cases, it is significantly more benficial
      to export the raw dataframe compared to a serialized version. In these cases, use
      ComponentIO.exportPkl. Otherwise, `pickle` is still a valid option here for a serialized
      format. For a full list of export options, see
      `the documentation`https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html`.
    :param pdExportArgs: Dictionary of values passed to underlying pandas export function.
      These will overwrite the default options for :func:`exportToFile <ComponentMgr.exportByFileType>`
    :param readOnly: Whether this export should be read-only
    :return: Export version of the component data.
    """
    # Make sure no rows are truncated
    col = None
    try:
      # TODO: Currently the additional options are causing errors. Find out why and fix
      #  them, since this may be useful if it can be modified
      # Format special columns appropriately
      # Since CSV export significantly modifies the df, make a copy before doing all these
      # operations
      exportDf = compDf.copy(deep=True)
      for col in exportDf:
        if not isinstance(col.value, str):
          serial, errs = serialize(col, exportDf[col])
          exportDf[col] = serial.to_numpy()
          if len(errs):
            raise ValueError(f'Encountered errors on the following rows:\n{errs.to_string()}')
      self._pandasSerialExport(exportDf, outFile, readOnly, **pdExportArgs)
    except Exception as ex:
      errMsg = f'Error on parsing column "{col.name}"\n'
      augmentException(ex, errMsg)
      raise
    return exportDf

  @wraps(exportSerialized, assigned=('__doc__', '__annotations__'))
  def exportCsv(self, *args, **kwargs):
    """Exposed format from the more general exportSerialized"""
    return self.exportSerialized(*args, **kwargs)

  @deprecateKwargs(imgDir='srcDir')
  def exportCompImgsDf(
      self,
      compDf: pd.DataFrame,
      outFile: Union[str, Path] = None,
      *,
      srcDir: Union[FilePath, dict, DirectoryDict] = None,
      margin=0,
      marginAsPct=False,
      includeCols=('instId', 'img', 'labelMask', 'label', 'offset'),
      lblField='Instance ID',
      asIndiv=False,
      returnLblMapping=False,
      missingOk=False,
      **kwargs
  ):
    """
    Creates a dataframe consisting of extracted images around each component
    :param compDf: Dataframe to export
    :param outFile: Where to save the result, if it should be saved. Caution -- this
      is currently a time-consuming process!
    :param srcDir: Where images corresponding to this dataframe are kept. Source image
      filenames are interpreted relative to this directory if they are not absolute. Alternatively, can be a dict
      of {name: np.ndarray} image mappings
    :param margin: How much padding to give around each component
    :param marginAsPct: Whether the margin should be a percentage of the component size or
      a raw pixel value.
    :param includeCols: Which columns to include in the export list
    :param lblField: See ComponentIO.exportLblPng. This label is provided in the output dataframe
      as well, if specified.
    :param asIndiv: Whether components should be exported as individual units (i.e.
      neighbors are guaranteed never to show up in the ground truth label mask) or whether
      a cropped area around the component should be used as the label mask. This will
      include mask values of neighbors if they are within mask range.
      Note: When false, the mask is created with preference toward higher ids, i.e. if a high ID is on
      top of a low ID, the low ID will still be covered in its export mask
    :param missingOk: Whether a missing image is acceptable. When no source image is found
      for an annotation, this will simpy the 'image' output property
    :param returnLblMapping: Whether to return the mapping of label numeric values to table field values
    :return: Dataframe with the following keys:
      - instId: The component's Instance ID
      - img: The (MxNxC) image corresponding to the component vertices, where MxN are
        the padded row sizes and C is the number of image channels
      - labelMask: Binary mask representing the component vertices
      - label: Field value of the component for the field specified by `lblField`
      - offset: Image (x,y) coordinate of the min component vertex.
    :param kwargs: Passed to ComponentIO.exportLblPng
    """
    bgColor = kwargs.get('bgColor', 0)
    _imgCache = {}
    if srcDir is None:
      srcDir = Path('.')
    srcDir = DirectoryDict(srcDir, allowAbsolute=True, readFunc=cvImread_rgb)
    uniqueImgs = np.unique(compDf[RTF.SRC_IMG_FILENAME])
    dfGroupingsByImg = []
    for imgName in uniqueImgs:
      dfGroupingsByImg.append(compDf[compDf[RTF.SRC_IMG_FILENAME] == imgName])

    useKeys = set(includeCols)
    outDf = {k: [] for k in useKeys}
    lblField = self.tableData.fieldFromName(lblField)
    # imshape is automatically inferred by the exporter
    kwargs.pop('imShape', None)
    mappings = {}
    for miniDf, fullImgName in zip(dfGroupingsByImg, uniqueImgs):
      img = srcDir.get(fullImgName)
      if img is None and not missingOk:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullImgName)
      shape = img if img is None else img.shape[:2]
      lblImg, mapping = self.exportLblPng(
          miniDf, imShape=shape, lblField=lblField, returnLblMapping=True, **kwargs
      )
      mappings[Path(fullImgName).name] = mapping
      if img is None:
        img = np.zeros_like(lblImg)
      invertedMap = pd.Series(mapping.index, mapping)

      for ii, (idx, row) in enumerate(miniDf.iterrows()):
        allVerts = row[RTF.VERTICES].stack()
        if marginAsPct:
          compImgSz = allVerts.max(0) - allVerts.min(0)
          marginToUse = (compImgSz * (margin/100)).astype(int)
        else:
          marginToUse = margin
        compImg, bounds = getCroppedImg(img, allVerts, marginToUse, coordsAsSlices=False)

        if 'img' in useKeys:
          outDf['img'].append(compImg)
        lbl = row[lblField]

        if 'label' in useKeys:
          outDf['label'].append(lbl)

        # x-y to row-col, transpose to get min-max in axis 0 and row-col in axis 1
        indexer = tuple(slice(*b) for b in bounds[:, ::-1].T)
        xyOffset = bounds[0, :]
        if lblImg.ndim > 2:
          indexer += (..., )
        if 'labelMask' in useKeys:
          mask = lblImg[indexer]
          if asIndiv:
            # Only color in this id's region. Since it might've been covered over by a different ID, regenerate the
            # mask from vertices, taking margin into account
            colorVerts = row[RTF.VERTICES].removeOffset(bounds[0, :])
            colorVerts.hierarchy = row[RTF.VERTICES].hierarchy
            mask = np.full_like(mask, bgColor)
            # 'Float' works for int and float orig values, required since opencv complains about some numpy
            # dtypes
            colorVerts.toMask(mask, float(invertedMap[lbl]), asBool=False)

          outDf['labelMask'].append(mask)

        if 'instId' in useKeys:
          outDf['instId'].append(idx)

        if 'offset' in useKeys:
          outDf['offset'].append(xyOffset)
    outDf = pd.DataFrame(outDf)
    if len(mappings) == 1:
      # Common case where annotations for just one image were converted
      mappings = next(iter(mappings.values()))
    outDf.attrs['mapping'] = mappings
    if outFile is not None:
      outDf.to_pickle(outFile)
    if returnLblMapping:
      return outDf, mappings
    return outDf

  def exportPkl(
      self,
      compDf: pd.DataFrame,
      outFile: Union[str, Path] = None,
      **exportArgs,
  ) -> (Any, str):
    """
    See the function signature for :func:`exportCsv <ComponentIO.exportCsv>`
    """
    # Since the write-out is a single operation there isn't an intermediate form to return
    pklDf = None
    if outFile is not None:
      pklDf = pickle.dumps(compDf)
      compDf.to_pickle(outFile)
    return pklDf

  def exportLblPng(
      self,
      compDf: pd.DataFrame,
      outFile: FilePath = None,
      imShape: Tuple[int] = None,
      lblField: Union[PrjParam, str] = 'Instance ID',
      bgColor=0,
      rescaleOutput=False,
      returnLblMapping=False,
      writeMeta=True,
      **kwargs
  ):
    """
    :param compDf: Dataframe to export
    :param outFile: Filename to save, leave *None* to avoid saving to a file
    :param imShape: MxN shape of image containing these annotations
    :param lblField: Data field to use as an index label. E.g. "Class" will use the 'class'
      column, but any other column can be specified. The output ground truth masks
      will be colored according to this field.  See :meth:`PrjParam.toNumeric` for details.
      If `lblField` is *None*, the foreground mask will be boolean instead of integer-colored.
    :param bgColor: Color of the mask background. Must be an integer.
    :param rescaleOutput: For images designed for human use, it is helpful to have
      outputs rescaled to the entire intensity range. Otherwise, they usually end
      up looking pure black and it is difficult to see components in the image.
      When `rescaleOutput` is *True*, all numbers are scaled to the 'uint16' range.
    :param returnLblMapping: Whether to return a pd.Series matching original index values
      to their numeric counterparts. Note: this is important in cases where an offset must be applied to the underlying
      data. If the background color is 0 and a valid numeric value is also 0, it will be impossible to detect this
      object in the labeled output. So, an offset must be applied in these cases (background - min(data) + 1). This
      mapping records the relevant information to import original values back during `importLblPng`.
    :param writeMeta: Whether to write the field mapping/offset to the output image file as png metadata.
      Useful to preserve label information when re-importing.
    :param kwargs:
    :return:
    """

    lblField = self.tableData.fieldFromName(lblField)

    if bgColor < 0:
      raise ValueError(f'Background color must be >= 0, was {bgColor}')

    readMapping = returnLblMapping or (writeMeta and outFile is not None)
    labels = compDf[lblField]
    labels_numeric = lblField.toNumeric(labels, returnMapping=readMapping)
    # Make sure numeric labels aren't the same as background, otherwise they will be forever lost
    mapping = None
    if readMapping:
      labels_numeric, mapping = labels_numeric
    diff = max(bgColor - np.min(labels_numeric, initial=0) + 1, 0)
    if readMapping:
      mapping.index += diff
    labels_numeric += diff
    asBool = np.issubdtype(labels_numeric.dtype, np.bool_)

    if rescaleOutput:
      if mapping is not None:
        max_ = np.max(np.asarray(mapping.index), initial=bgColor)
        mapping.index = rescale_intensity(
            mapping.index, in_range=(bgColor, max_), out_range='uint16'
        )
      else:
        max_ = np.max(labels_numeric, initial=bgColor)
      labels_numeric = rescale_intensity(
          labels_numeric, in_range=(bgColor, max_), out_range='uint16'
      )

    if imShape is None:
      # Without any components the image is non-existant
      if len(compDf) == 0:
        raise ValueError('imShape cannot be *None* if no components are present')
      vertMax = ComplexXYVertices.stackedMax(compDf[RTF.VERTICES])
      imShape = tuple(vertMax[::-1] + 1)
    maskType = 'uint16' if np.min(bgColor) >= 0 else 'int32'
    outMask = np.full(imShape[:2], bgColor, dtype=maskType)
    for fillClr, (_, comp) in zip(labels_numeric, compDf.iterrows()):
      verts: ComplexXYVertices = comp[RTF.VERTICES]
      outMask = verts.toMask(outMask, int(fillClr), False, False)
    if asBool:
      outMask = outMask > 0

    if outFile is not None:
      if writeMeta:
        _writeImge_meta(outMask, outFile, mapping)
      else:
        cvImsave_rgb(outFile, outMask)
    if returnLblMapping:
      return outMask, mapping
    return outMask

  # def exportClassPng(self, compDf: pd.DataFrame, outFile: FilePath = None, imShape: Tuple[int]=None, **kwargs):
  #   # Create label to output mapping
  #   return self.exportLblPng(compDf, outFile, imShape, 'Class', **kwargs)

  def exportCompImgsZip(
      self,
      compDf: pd.DataFrame,
      outDir: FilePath = 's3a-export',
      resizeShape: Tuple[int, int] = None,
      archive=False,
      makeSummary=False,
      **kwargs
  ):
    """
    From a component dataframe, creates output directories for component images and masks.
    This is useful for many neural networks etc. to read individual component images.

    :param compDf: Dataframe to export
    :param outDir:
      helpText: "Where to make the output directories. If `None`, defaults to current
      directory>compimgs_<margin>_margin"
      pType: filepicker
    :param resizeShape: If provided, it is the shape that all images will be resized to before
      being saved. This is useful for neural networks with a fixed input size which forces all
      inputs to be e.g. 100x100 pixels.
    :param archive: Whether to compress into a zip archive instead of directly outputting a folder
    :param makeSummary: Whether to include an html table showing each component from the dataframe along with
      its image and mask representations
    :param kwargs: Passed directly to :meth:`ComponentIO.exportCompImgsDf`
    """
    outDir = Path(outDir)
    useDir = outDir

    saveFn = lambda fname, img: cvImsave_rgb(fname, img)
    if resizeShape is not None:
      saveFn = lambda fname, img: cvImsave_rgb(fname, resize_pad(img, resizeShape))

    with ExitStack() as stack:
      if archive:
        useDir = Path(stack.enter_context(tempfile.TemporaryDirectory()))
      dataDir = useDir / 'data'
      labelsDir = useDir / 'labels'
      dataDir.mkdir(exist_ok=True, parents=True)
      labelsDir.mkdir(exist_ok=True, parents=True)

      summaryName = useDir / 'summary.html'

      extractedImgs = self.exportCompImgsDf(compDf, None, **kwargs)
      for idx, row in extractedImgs.iterrows():
        saveName = f'{row.instId}.png'
        if 'img' in row.index:
          saveFn(dataDir / saveName, row.img)
        if 'labelMask' in row.index:
          saveFn(labelsDir / saveName, row.labelMask)

      if makeSummary:
        extractedImgs = extractedImgs.rename({'instId': RTF.INST_ID.name}, axis=1)
        # Prevent merge error by renaming index
        # INST_ID.name has to be used instead of raw INST_ID due to strange pandas issue
        # throwing a TypeError: keywords must be a string
        outDf: pd.DataFrame = compDf.drop([RTF.VERTICES], axis=1).rename(str, axis=1)
        outDf = outDf.merge(
          extractedImgs, on=RTF.INST_ID.name
        )
        for colName, imgDir in zip(['labelMask', 'img'], [labelsDir, dataDir]):
          if colName not in extractedImgs:
            continue
          relDir = imgDir.relative_to(useDir)
          outDf[colName] = outDf[RTF.INST_ID.name].apply(
              lambda el: imgPathtoHtml((relDir / str(el)).with_suffix('.png').as_posix())
          )
        outDf.columns = list(map(str, outDf.columns))
        outDf.to_html(summaryName, escape=False, index=False)

      if archive:
        if outDir.suffix != '.zip':
          outDir = outDir.with_suffix(outDir.suffix + '.zip')
        with ZipFile(outDir, 'w') as ozip:
          for dir_ in labelsDir, dataDir:
            if not dir_.exists():
              continue
            for file in dir_.iterdir():
              ozip.write(file, f'{dir_.name}/{file.name}')
          if makeSummary:
            ozip.write(summaryName, file.name)

  # -----
  # Import options
  # -----
  def convert(
      self,
      fromData: FilePathOrDf,
      toFile: FilePath,
      doExport=True,
      importArgs: dict = None,
      exportArgs: dict = None
  ):
    if importArgs is None:
      importArgs = {}
    if exportArgs is None:
      exportArgs = {}
    if not isinstance(fromData, pd.DataFrame):
      fromData = self.importByFileType(fromData, **importArgs)
    exportFn = self._ioFnFromFileType(toFile, PRJ_ENUMS.IO_EXPORT)
    if not doExport:
      toFile = None
    return exportFn(fromData, toFile, **exportArgs)

  @fns.dynamicDocstring(availImporters=_getPdImporters())
  def importSerialized(
      self,
      inFileOrDf: FilePathOrDf,
      imShape: Tuple = None,
      reindex=False,
      parseErrorOk=False,
      **importArgs
  ):
    """
    Deserializes data from a file or string dataframe to create a S3A Component
    :class:`DataFrame`.
    The input file or dataframe should be the same format as one exported by
    :func:`exportSerialized <ComponentMgr.exportSerialized>`.

    :param imShape: If included, this ensures all imported components lie within imSize
           boundaries. If any components do not, an error is thrown since this is
           indicative of components that actually came from a different reference image.
    :param inFileOrDf: Name of file to import, or dataframe itself if it was already read
    from this file. Note: avoid passing in a dataframe directly unless you know what
    you're doing since null values and some converter types can produce misleading
    values. Most file types supported by pandas (pandas.read_*) are supported here.
    :param reindex: Whether to disregard the index of the incoming dataframe or file.
      This is useful when *inFileOrDf* is actually a conacatenated df of multiple files, and
      the index doesn't need to be retained.
    :return: Tuple: pd.DataFrame that will be exported if successful extraction
    """
    fileName = None if isinstance(inFileOrDf, pd.DataFrame) else inFileOrDf
    field = PrjParam('None', None)
    serialDf = self._pandasSerialImport(inFileOrDf, **importArgs)
    totalErrs = pd.DataFrame()
    if reindex:
      serialDf[RTF.INST_ID.name] = np.arange(len(serialDf), dtype=int)
      serialDf = serialDf.set_index(RTF.INST_ID.name, drop=False)
    try:
      # Decouple index from instance ID until after transfer from csvDf is complete
      # This was causing very strange behavior without reset_index()...
      outDf = self.tableData.makeCompDf(len(serialDf)).reset_index(drop=True)
      # Objects in the original frame are repre sented as strings, so try to convert these
      # as needed
      for field in self.tableData.allFields:
        errs = []
        if field.name in serialDf:
          matchingCol = serialDf[field.name]
          # 'Object' type results in false positives
          # if matchingCol.dtype != object and type(field.value) == matchingCol.dtype:
          #   outDf[field] = serialDf[field.name]
          # else:
          # Parsing functions only know how to convert from strings to themselves.
          # So, assume the exting types can first convert themselves to strings
          matchingCol = matchingCol.astype(str)
          outDf[field], errs = deserialize(field, matchingCol)
          if len(errs):
            totalErrs = totalErrs.join(errs, how='outer')
      if len(totalErrs) and not parseErrorOk:
        raise AnnParseError(
            f'Encountered problems on some annotations:\n{totalErrs.to_string()}',
            fileName=fileName,
            instances=totalErrs
        )
      outDf = outDf.set_index(RTF.INST_ID, drop=False)

      self.checkVertBounds(outDf[RTF.VERTICES], imShape)
    except Exception as ex:
      # Rethrow exception with insight about column number
      # Procedure copied from https://stackoverflow.com/a/6062677/9463643
      errMsg = f'Error importing column "{field.name}":\n'
      augmentException(ex, errMsg)
      raise
    # TODO: Apply this function to individual rows instead of the whole dataframe. This will allow malformed
    #  rows to gracefully fall off the dataframe with some sort of warning message
    return outDf

  @wraps(importSerialized, assigned=('__doc__', '__annotations__'))
  def importCsv(self, *args, **kwargs):
    """Exposed format from the more general importSerialized"""
    return self.importSerialized(*args, **kwargs)

  def _foreignSerialImport(
      self,
      foreignTbl: TableData,
      foreignDf: pd.DataFrame,
      mapping: dict = None,
  ):
    """
    Several forms of imports / exports handle data that may not be compatible with the current table data.
    In these cases, it is beneficial to determine a mapping between names to allow greater compatibility between
    I/O formats. This also safely attempts a serial transfer without leaving self's table data in an unstable state.
    Mapping is also extended in both directions by parameter name aliases (param.opts['aliases']), which are a list
    of strings of common mappings for that parameter (e.g. [Class, Label] are often used interchangeably)

    :param foreignTbl: Table of foreign information
    :param foreignDf: Dataframe with maybe foreign fields
    :param mapping: Foreign to local field name mapping
    """
    oldTbl = self.tableData
    try:
      self.tableData = foreignTbl
      foreignDf = self.importSerialized(foreignDf)
    finally:
      self.tableData = oldTbl
    outCols = list(foreignTbl.allFields)

    for ii, field in enumerate(foreignTbl.allFields):
      checkNames = set([field.name] + field.opts.get('aliases', []))
      for cmp in oldTbl.allFields:
        cmpNames = [cmp.name] + cmp.opts.get('aliases', [])
        if checkNames & set(cmpNames):
          # There's a match!
          outCols[ii] = cmp

    if mapping:
      # Mapping takes priority, so let it overwrite potentially alias-mapped values
      for kk, vv in mapping.items():
        if vv in oldTbl and kk in outCols:
          outCols[kk] = vv
    foreignDf.columns = outCols
    return foreignDf

  def importGeojson(
      self,
      inFileOrDict: Union[FilePath, dict],
      parseErrorOk=False,
      **importArgs,
  ):
    fileName = None
    if not isinstance(inFileOrDict, dict):
      fileName = Path(inFileOrDict)
      with open(inFileOrDict, 'r') as ifile:
        inFileOrDict = json.load(ifile)
    verts = []
    badInsts = []
    for ann in inFileOrDict['features']:
      geo = ann['geometry']
      if geo['type'] == 'Polygon':
        verts.append(ComplexXYVertices(geo['coordinates'], coerceListElements=True))
      else:
        badInsts.append(ann)
    if badInsts and not parseErrorOk:
      raise AnnParseError(
          f'Currently, S3A only supports polygon geojson types',
          fileName=fileName,
          instances=badInsts
      )
    outDf = self.tableData.makeCompDf(len(verts))
    outDf[RTF.VERTICES] = verts
    return outDf

  def importSuperannotateJson(
      self,
      inFileOrDict: Union[FilePath, dict],
      parseErrorOk=False,
      srcDir: Union[FilePath, dict] = None,
      **importArgs,
  ):
    """
    :param inFileOrDict: File or object to import
    :param parseErrorOk: Whether errors should be raised on bad annotation instances
    :param srcDir: Folder path with 'classes.json' (usually this folder is called 'classes')
      or dict with 'classes.json' key and dict value
    """
    fileName = None

    def readFunc(file):
      with open(file, 'r') as ifile:
        return json.load(ifile)

    if not isinstance(inFileOrDict, dict):
      fileName = Path(inFileOrDict)
      inFileOrDict = readFunc(inFileOrDict)
    foreignDf = pd.DataFrame(inFileOrDict['instances'])
    if srcDir is None:
      srcDir = Path(f'.')
    srcDir = DirectoryDict(srcDir, readFunc=readFunc, allowAbsolute=True)
    useTbl = self._getForeignTableData('superannotateJson')
    classes = srcDir.get('classes.json')
    if classes is None and fileName is not None:
      classes = srcDir.get(fileName.parent / 'classes' / 'classes.json')
    if classes is not None:
      useTbl.fieldFromName('className').opts['limits'] = [c['name'] for c in classes]
    instances = inFileOrDict['instances']
    parsePts = []
    invalidInsts = []
    invalidIdxs = []
    for ii, inst in enumerate(instances):
      typ = inst['type']
      if typ in ['polygon', 'bbox']:
        pts = inst['points']
        if typ == 'bbox':
          pts = list(pts.values())
        parsePts.append(np.column_stack([pts[::2], pts[1::2]]))
      elif typ == 'ellipse':
        vals = inst['cy'], inst['cx'], inst['ry'], inst['rx'], inst['angle']
        pts = draw.ellipse_perimeter(*(int(v) for v in vals))
        pts = np.column_stack(pts[::-1])
        parsePts.append(orderContourPts(pts))
      else:
        invalidInsts.append(inst)
        invalidIdxs.append(ii)
    if invalidInsts and not parseErrorOk:
      raise AnnParseError(
          'Currently, S3A only supports polygon annotations from SuperAnnotate',
          fileName=fileName,
          instances=invalidInsts
      )
    foreignDf = foreignDf.drop(invalidIdxs)
    outDf = self._foreignSerialImport(useTbl, foreignDf, importArgs.get('mapping'))

    outDf[RTF.VERTICES] = [ComplexXYVertices([pts], coerceListElements=True) for pts in parsePts]
    outDf[RTF.SRC_IMG_FILENAME] = inFileOrDict['metadata']['name']
    imShape = (inFileOrDict['metadata']['height'], inFileOrDict['metadata']['width'])
    self.checkVertBounds(outDf[RTF.VERTICES], imShape)
    return outDf

  def importCompImgsDf(
      self,
      inFile: FilePath,
      imShape: Tuple = None,
      lblField='Instance ID',
      **importArgs,
  ):
    lblField = self.tableData.fieldFromName(lblField)
    inDf = pd.read_pickle(inFile)
    outDf = self.tableData.makeCompDf(len(inDf))
    outDf[RTF.INST_ID] = inDf['instId']
    allVerts = []

    for idx, row in inDf.iterrows():
      mask = row.labelMask
      verts = ComplexXYVertices.fromBwMask(mask).remove(-row.offset)
      allVerts.append(verts)
    outDf[RTF.VERTICES] = allVerts
    outDf[lblField] = inDf['label']
    self.checkVertBounds(outDf[RTF.VERTICES], imShape)
    return outDf

  def importPkl(self, inFile: FilePath, imShape: Tuple = None, **importArgs) -> pd.DataFrame:
    """
    See docstring for :func:`self.importCsv`
    """
    pklDf = pd.read_pickle(inFile)
    self.checkVertBounds(pklDf[RTF.VERTICES], imShape)
    templateDf = self.tableData.makeCompDf(len(pklDf)).set_index(pklDf.index)
    templateDf.update(pklDf)
    return templateDf

  def importLblPng(
      self,
      inFileOrImg: Union[FilePath, GrayImg],
      lblField: Union[str, PrjParam] = None,
      lblMapping: pd.Series = None,
      distinctRegions=True,
      offset=0,
      **importArgs
  ) -> pd.DataFrame:
    """
    Build dataframe from label image

    :param inFileOrImg: Image or image file to parse
    :param lblField: label field to associate with this image. Pixels values within the image
      correspond to values from this field in the table data. If *None*, this is inferred by the mapping read
      from the image file (see `lblMapping` description)
    :param lblMapping: For parameters that aren't numeric and don't have limits (e.g. arbitrary string values),
      this mapping determines how numeric values should be turned into field values. See `PrjParam.toNumeric` for
      details, since this is the mapping expected. If not provided, first the image metadata tags are searched for
      a 'lblMapping' text attribute (this is often added to label images saved by S3A). Note that metadata can only be
      read from the file if a file path is provided, of course. If this check fails, it is inferred based on the
      allowed options of `lblField` (`lblField.opts['limits']`). Finally, if this is not present, it is assumed the
      raw image values can be used directly as field values.
    :param offset: When `lblMapping` is not provided and field values are directly inferred from label values, this
      determines whether (and how much if not *None*) to offset numeric labels during import. I.e. if the png label
      is 1, but offset is 1, the corresponding *field* value will be 0 (1 - offset = 0).
    :param distinctRegions: Whether separate regions with the same ID should be separate IDs, or
      one ID with a group of polygons
    """
    if lblMapping is None:
      # Ignoring offset for now
      labelImg, lblMapping, _ = _getImg_mapping_offset(inFileOrImg)
    else:
      labelImg = cvImread_rgb(str(inFileOrImg), mode=cv.IMREAD_UNCHANGED)
    lblField = self.tableData.fieldFromName(lblField or lblMapping.name)
    if lblMapping is None:
      vals = lblField.opts.get('limits', None) or np.unique(labelImg)
      _, lblMapping = lblField.toNumeric(vals, returnMapping=True)
      lblMapping.index += offset
    allVerts = []
    lblField_out = []
    for numericLbl, origVal in lblMapping.iteritems():
      origVal: np.number  # silence warning
      verts = ComplexXYVertices.fromBwMask(labelImg == numericLbl)
      if distinctRegions:
        newRegions = [ComplexXYVertices([v]) for v in verts]
        allVerts.extend(newRegions)
        orig = np.tile(origVal, len(verts))
      else:
        allVerts.append(verts)
        orig = [origVal]
      lblField_out.extend(orig)
    outDf = self.tableData.makeCompDf(len(allVerts))
    outDf[lblField] = lblField_out
    outDf[RTF.VERTICES] = allVerts
    return outDf

  @staticmethod
  def checkVertBounds(vertSer: pd.Series, imShape: tuple):
    """
    Checks whether any vertices in the imported dataframe extend past image dimensions. This is an indicator
    they came from the wrong import file.

    :param vertSer: Vertices from incoming component dataframe
    :param imShape: Shape of the main image these vertices are drawn on
    :return: Raises error if offending vertices are present, since this is an indication the component file
      was from a different image
    """
    if imShape is None or len(vertSer) == 0:
      # Nothing we can do if no shape is given
      return
    # Image shape from row-col -> x-y
    imShape = np.array(imShape[1::-1])[None, :]
    # Remove components whose vertices go over any image edges
    vertMaxs = [verts.stack().max(0) for verts in vertSer if len(verts) > 0]
    vertMaxs = np.vstack(vertMaxs)
    offendingIds = np.nonzero(np.any(vertMaxs > imShape, axis=1))[0]
    if len(offendingIds) > 0:
      warnLater(
          f'Vertices on some components extend beyond image dimensions. '
          f'Perhaps this export came from a different image?\n'
          f'Offending IDs: {offendingIds}', UserWarning
      )

  def _idImgToDf(self, idImg: GrayImg):
    # Skip 0 since it's indicative of background
    regionIds = np.unique(idImg)
    regionIds = regionIds[regionIds != 0]
    allVerts = []
    for curId in regionIds:
      verts = ComplexXYVertices.fromBwMask(idImg == curId)
      allVerts.append(verts)
    outDf = self.tableData.makeCompDf(regionIds.size)
    # Subtract 1 since instance ids are 0-indexed
    outDf[RTF.INST_ID] = regionIds - 1
    outDf[RTF.VERTICES] = allVerts
    return outDf


defaultIo = ComponentIO()



import html
import warnings
from collections import defaultdict
from functools import wraps
from pathlib import Path
from typing import Callable, Tuple, Union, Sequence, List, Collection, Any

import cv2 as cv
import numpy as np
import pyqtgraph as pg
from pandas import DataFrame as df
from skimage import io
from skimage import transform
from skimage.exposure import exposure

from utilitys import PrjParam, fns, ProcessStage
# Needs to be visible outside this file
# noinspection PyUnresolvedReferences
from utilitys.fns import hierarchicalUpdate
from utilitys.typeoverloads import FilePath
from .structures import TwoDArr, XYVertices, ComplexXYVertices, NChanImg, BlackWhiteImg


def stackedVertsPlusConnections(vertList: ComplexXYVertices) -> (XYVertices, np.ndarray):
  """
  Utility for concatenating all vertices within a list while recording where separations
  occurred
  """
  allVerts = [np.zeros((0,2))]
  separationIdxs = []
  idxOffset = 0
  for curVerts in vertList:
    allVerts.append(curVerts)
    vertsLen = len(curVerts)
    if vertsLen == 0: continue
    # Close the current shape
    allVerts.append(curVerts[0,:])
    separationIdxs.append(idxOffset + vertsLen)
    idxOffset += vertsLen + 1
  # Take away last separator if it exists
  if len(separationIdxs) > 0:
    separationIdxs.pop()
  allVerts = np.vstack(allVerts)
  isfinite = np.ones(len(allVerts), bool)
  isfinite[separationIdxs] = False
  return XYVertices(allVerts, dtype=float), isfinite
  #return XYVertices(dtype=float)

def getClippedBbox(arrShape: tuple, bbox: TwoDArr, margin: int):
  """
  Given a bounding box and margin, create a clipped bounding box that does not extend
  past any dimension size from arrShape

  Parameters
  ----------
  arrShape :    2-element tuple
     Refrence array dimensions

  bbox     :    2x2 array
     [minX minY; maxX maxY] bounding box coordinates

  margin   :    int
     Offset from bounding box coords. This will not fully be added to the bounding box
     if the new margin causes coordinates to fall off either end of the reference array shape.
  """
  bbox = bbox.astype(int)
  bbox[0] -= margin
  # Add extra 1 since python slicing stops before last index
  bbox[1] += margin+1
  arrShape = arrShape[:2]
  return np.clip(bbox, 0, arrShape[::-1])

def coerceDfTypes(dataframe: df, constParams: Collection[PrjParam]=None):
  """
  Pandas currently has a bug where datatypes are not preserved after update operations.
  Current workaround is to coerce all types to their original values after each operation
  """
  if constParams is None:
    constParams = dataframe.columns
  for field in constParams:
    try:
      dataframe[field] = dataframe[field].astype(type(field.value))
    except TypeError:
      # Coercion isn't possible, nothing to do here
      pass
  return dataframe

def largestList(verts: List[XYVertices]) -> XYVertices:
  maxLenList = []
  for vertList in verts:
    if len(vertList) > len(maxLenList): maxLenList = vertList
  # for vertList in newVerts:
  # vertList += cropOffset[0:2]
  return XYVertices(maxLenList)

def augmentException(ex: Exception, prependedMsg: str):
  exMsg = str(ex)
  ex.args = (prependedMsg + exMsg,)

def lower_NoSpaces(name: str):
  return name.replace(' ', '').lower()

def safeCallFuncList(fnNames: Collection[str], funcLst: List[Callable],
                     fnArgs: List[Sequence]=None):
  errs = []
  rets = []
  if fnArgs is None:
    fnArgs = [()]*len(fnNames)
  for key, fn, args in zip(fnNames, funcLst, fnArgs):
    curRet, curErr = safeCallFunc(key, fn, *args)
    rets.append(curRet)
    if curErr: errs.append(curErr)
  return rets, errs

def safeCallFunc(fnName: str, func: Callable, *fnArgs):
  ret = err = None
  try:
    ret = func(*fnArgs)
  except Exception as ex:
    err = f'{fnName}: {ex}'
  return ret, err

def _maybeBgrToRgb(image: np.ndarray):
  """Treats 3/4-channel images as BGR/BGRA for opencv saving/reading"""
  if image.ndim > 2:
    # if image.shape[0] == 1:
    #   image = image[...,0]
    if image.shape[2] >= 3:
      lastAx = np.arange(image.shape[2], dtype='int')
      # Swap B & R
      lastAx[[0,2]] = [2,0]
      image = image[...,lastAx]
  return image

def cvImsave_rgb(fname: FilePath, image: np.ndarray, *args, errOk=False, **kwargs):
  image = _maybeBgrToRgb(image)
  try:
    cv.imwrite(str(fname), image, *args, **kwargs)
  except cv.error:
    if not errOk:
      raise
    # Dtype incompatible
    io.imsave(fname, image)

def cvImread_rgb(fname: FilePath, *args, **kwargs):
  image = cv.imread(str(fname), *args, **kwargs)
  return _maybeBgrToRgb(image)

def tryCvResize(image: NChanImg, newSize: Union[tuple, float],
                asRatio=True, interp=cv.INTER_CUBIC):
  """
  Uses cv.resize where posisble, but if dtypes prevent this, it falls back to skimage.transform.rescale/resize

  :param image: Image to resize
  :param newSize: Either ratio to scale each axis or new image size (x, y -- not row, col)
  :param asRatio: Whether to interpret `newSize` as a ratio or new image dimensions
  :param interp: Interpolation to use, if cv.resize is available for the given dtype
  """
  if asRatio:
    if not isinstance(newSize, tuple):
      newSize = (newSize, newSize)
    args = dict(dsize=(0,0), fx=newSize[0], fy=newSize[1])
  else:
    args = dict(dsize=newSize)
  try:
    image = cv.resize(image, **args, interpolation=interp)
  except (TypeError, cv.error):
    oldRange = (image.min(), image.max())
    if asRatio:
      rescaled = transform.rescale(image, newSize, multichannel=image.ndim > 2)
    else:
      rescaled = transform.resize(image, newSize[::-1])
    image = exposure.rescale_intensity(rescaled, out_range=oldRange).astype(image.dtype)
  return image



def cornersToFullBoundary(cornerVerts: Union[XYVertices, ComplexXYVertices], sizeLimit: float=np.inf,
                          fillShape: Tuple[int]=None, stackResult=True) -> Union[XYVertices, ComplexXYVertices]:
  """
  From a list of corner vertices, returns a list with one vertex for every border pixel.
  Example:
  >>> cornerVerts = XYVertices([[0,0], [100,0], [100,100],[0,100]])
  >>> cornersToFullBoundary(cornerVerts)
  # [[0,0], [1,0], ..., [100,0], [100,1], ..., [100,100], ..., ..., [0,100]]
  :param cornerVerts: Corners of the represented polygon
  :param sizeLimit: The largest number of pixels from the enclosed area allowed before the full boundary is no
  longer returned. For instance:
    >>> cornerVerts = XYVertices([[0,0], [1000,0], [1000,1000],[0,1000]])
    >>> cornersToFullBoundary(cornerVerts, 10e5)
    will *NOT* return all boundary vertices, since the enclosed area (10e6) is larger than sizeLimit.
  :param fillShape: Size of mask to create. Useful if verts may extend beyond image dimensions
    and should be truncated. If None, no truncation will occur except for negative verts.
  :param stackResult: Whether the result should be ComplexXYVertices (if stackResult is False)
    or a stacked list of exterior verts (if stackResult is True)
  :return: List with one vertex for every border pixel, unless *sizeLimit* is violated.
  """
  if isinstance(cornerVerts, XYVertices):
    cornerVerts = ComplexXYVertices([cornerVerts])
  if fillShape is not None:
    fillShape = tuple(fillShape)
  filledMask = cornerVerts.toMask(fillShape, warnIfTooSmall=False)
  cornerVerts = ComplexXYVertices.fromBwMask(filledMask, simplifyVerts=False)
  if not stackResult:
    return cornerVerts
  cornerVerts = cornerVerts.filledVerts().stack()
  numCornerVerts = len(cornerVerts)
  if numCornerVerts > sizeLimit:
    spacingPerSamp = int(numCornerVerts/sizeLimit)
    cornerVerts = cornerVerts[::spacingPerSamp]
  return cornerVerts


def getCroppedImg(image: NChanImg, verts: np.ndarray, margin=0, *otherBboxes: np.ndarray,
                  coordsAsSlices=False, returnCoords=True) -> (np.ndarray, np.ndarray):
  """
  Crops an image according to the specified vertices such that the returned image does not extend
  past vertices plus margin (including other bboxes if specified). All bboxes and output coords
  are of the form [[xmin, ymin], [xmax, ymax]]. Slices are (row slices, col slices) if `coordsAsSlices`
  is specified.
  """
  verts = np.vstack(verts)
  img_np = image
  compCoords = np.vstack([verts.min(0), verts.max(0)])
  if len(otherBboxes) > 0:
    for dim in range(2):
      for ii, cmpFunc in zip(range(2), [min, max]):
        otherCmpVals = [curBbox[ii, dim] for curBbox in otherBboxes]
        compCoords[ii,dim] = cmpFunc(compCoords[ii,dim], *otherCmpVals)
  compCoords = getClippedBbox(img_np.shape, compCoords, margin)
  coordSlices = (slice(compCoords[0,1], compCoords[1,1]),
                 slice(compCoords[0,0],compCoords[1,0]))
  # Verts are x-y, index into image with row-col
  indexer = coordSlices
  if image.ndim > 2:
    indexer += (slice(None),)
  croppedImg = image[indexer]
  if not returnCoords:
    return croppedImg
  if coordsAsSlices:
    return croppedImg, coordSlices
  else:
    return croppedImg, compCoords

def imgCornerVertices(img: NChanImg=None) -> XYVertices:
  """Returns [x,y] vertices for each corner of the input image"""
  if img is None:
    return XYVertices()
  fullImShape_xy = img.shape[:2][::-1]
  return XYVertices([[0,                   0],
              [0,                   fullImShape_xy[1]-1],
              [fullImShape_xy[0]-1, fullImShape_xy[1]-1],
              [fullImShape_xy[0]-1, 0]
              ])

def resize_pad(img: NChanImg, newSize: Tuple[int, int], interp=cv.INTER_NEAREST, padVal=0):
  """
  Resizes image to the requested size using the specified interpolation method.
  :param img: Image to resize
  :param newSize: New size for image. Since aspect ratio is maintained, the portion of the
    image which couldn't be resized fully is padded with a constant value of `padVal`.
    For instance, if the original image is 5x10 and the requested new size is 10x15, then
    after resizing the image will be 7x15 to preserve aspect ratio. 2 pixels of padding
    will be added on the left and 1 pixel of padding will be added on the right so the final
    output is 10x15.
  :param padVal: Value to pad dimension that couldn't be fully resized
  :param interp: Interpolation method to use during resizing
  :param padVal: Constant value to use during padding
  :return: Resized and padded image
  """
  needsRotate = False
  # Make sure largest dimension is along rows for easier treatment and rotate back after
  newSize = np.array(newSize)
  resizeRatio = min(newSize/np.array(img.shape[:2]))
  paddedImg = cv.resize(img, (0, 0), fx = resizeRatio, fy = resizeRatio, interpolation=interp)
  padDimension = np.argmax(newSize - np.array(paddedImg.shape[:2]))
  if padDimension != 1:
    paddedImg = cv.rotate(paddedImg, cv.ROTATE_90_CLOCKWISE)
    newSize = newSize[::-1]
    needsRotate = True
  # Now pad dimension is guaranteed to be index 1
  padDimension = 1
  padding = (newSize[padDimension] - paddedImg.shape[padDimension]) // 2
  if padding > 0:
    paddedImg = cv.copyMakeBorder(paddedImg, 0, 0, padding, padding, cv.BORDER_CONSTANT, value=padVal)
  # Happens during off-by-one truncated division
  remainingPadding = newSize[padDimension] - paddedImg.shape[padDimension]
  if remainingPadding > 0:
    paddedImg = cv.copyMakeBorder(paddedImg, 0, 0, remainingPadding, 0, cv.BORDER_CONSTANT, value=padVal)
  if needsRotate:
    paddedImg = cv.rotate(paddedImg, cv.ROTATE_90_COUNTERCLOCKWISE)
  return paddedImg

def showMaskDiff(oldMask: BlackWhiteImg, newMask: BlackWhiteImg):
  infoMask = np.tile(oldMask[...,None].astype('uint8')*255, (1,1,3))
  # Was there, now it's not -- color red
  infoMask[oldMask & ~newMask,:] = [255,0,0]
  # Wasn't there, now it is -- color green
  infoMask[~oldMask & newMask,:] = [0,255,0]
  return infoMask

# Poor man's LRU dict, since I don't yet feel like including another pypi dependency
class MaxSizeDict(dict):
  def __init__(self, *args, maxsize:int=np.inf, **kwargs):
    super().__init__(*args, **kwargs)
    self.maxsize = maxsize

  def __setitem__(self, key, value):
    if len(self) >= self.maxsize:
      # Evict oldest inserted entry
      self.pop(next(iter(self.keys())))
    super().__setitem__(key, value)

def _getPtAngles(pts):
  midpt = np.mean(pts, 0)
  relPosPts = (pts - midpt).view(np.ndarray)
  return np.arctan2(*relPosPts.T[::-1])

def orderContourPts(pts: XYVertices, ccw=True):
  """
  Only guaranteed to work for convex hulls, i.e. shapes not intersecting themselves. Orderes
  an arbitrary list of coordinates into one that works well line plotting, i.e. doesn't show
  excessive self-crossing when plotting
  """
  if len(pts) < 3:
    return pts
  angles = _getPtAngles(pts)
  ptOrder = np.argsort(angles)
  if not ccw:
    ptOrder = ptOrder[::-1]
  return pts[ptOrder]

# def movePtsTowardCenter(pts: XYVertices, dist=1):
#   if not pts.size:
#     return pts
#   angles = _getPtAngles(pts)
#   adjusts = np.column_stack([np.cos(angles), np.sin(angles)])
#   adjusts[np.abs(adjusts) < 0.01] = 0
#   # Adjust by whole steps
#   adjusts = np.sign(adjusts)*dist
#   return pts - adjusts

def symbolFromVerts(verts: ComplexXYVertices):
  concatRegion, isfinite = stackedVertsPlusConnections(verts)
  if not len(concatRegion):
    boundLoc = np.array([[0,0]])
  else:
    boundLoc = np.nanmin(concatRegion, 0, keepdims=True)
  useVerts = concatRegion - boundLoc + 0.5
  return pg.arrayToQPath(*useVerts.T, connect=isfinite), boundLoc

# Credit: https://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
def bboxIou(boxA, boxB):
  """
  determine the (x, y)-coordinates of the intersection rectangle. Both boxes are formatted
   [[xmin, ymin], [xmax, ymax]]"""
  boxA = boxA.ravel()
  boxB = boxB.ravel()
  xA = max(boxA[0], boxB[0])
  yA = max(boxA[1], boxB[1])
  xB = min(boxA[2], boxB[2])
  yB = min(boxA[3], boxB[3])
  # compute the area of intersection rectangle
  interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
  # compute the area of both the prediction and ground-truth
  # rectangles
  boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
  boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)
  # compute the intersection over union by taking the intersection
  # area and dividing it by the sum of prediction + ground-truth
  # areas - the interesection area
  iou = interArea / float(boxAArea + boxBArea - interArea)
  # return the intersection over union value
  return iou

# Credit: https://stackoverflow.com/a/13624858/9463643
class classproperty:

  def __init__(self, fget):
    self.fget = fget

  def __get__(self, owner_self, owner_cls):
    return self.fget(owner_cls)

def imgPathtoHtml(imgPath: FilePath, width=None):
  outStr = f'<img src="{imgPath}"'
  if width is not None:
    outStr += f' width="{width}px"'
  outStr += '>'
  return outStr

def incrStageNames(stages: Sequence[ProcessStage]):
  """
  Returns a list of names numerically incrementing where they would otherwise be duplicated. E.g. if the stage
  list was [Open, Open], this is converted to [Open, Open#2]
  """
  preExisting = defaultdict(int)
  names = []
  for stage in stages:
    name = stage.name
    if preExisting[name] > 0:
      name += f'#{preExisting[name] + 1}'
    preExisting[name] += 1
    names.append(name)
  return names

class DirectoryDict(MaxSizeDict):
  """
  Used to shim the API between file-system and programmatically generated content. If a directory is passed, files are
  read and cached when a name is passed. Otherwise, treat as a normal dict of things. For instance, a directory of
  png images can be accessed through this data structure as ddict['x.png'] which will load the image. Next time 'x.png'
  is accessed, it will be instantaneous due to caching. Similarly, `ddict` could be given an initial dict if contents
  are not directory-backed
  """
  _UNSET = object()

  def __init__(self, initData:Union[FilePath, dict, 'DirectoryDict']=None, readFunc: Callable[[str], Any]=None, allowAbsolute=False,
               **kwargs,):
    """
    :param initData: Either startup dict or backing directory path. If a DirectoryDict is passed, its attribute will
      be used instead of the value passed for allowAbsolute. Its readFunc will be used if the passed
      readFunc is *None*
    :param readFunc: Function used to read files from the directory, i.e. `io.imread`, `attemptFileLoad`, etc.
      Must accept the name of the file to read
    :param allowAbsolute: Whether to allow reading absolute paths
    :param kwargs: Passed to super constructor
    """
    if isinstance(initData, FilePath.__args__):
      self.fileDir = Path(initData)
      super().__init__(**kwargs)
    else:
      if isinstance(initData, DirectoryDict):
        readFunc = readFunc or initData.readFunc
        allowAbsolute = initData.allowAbsolute
        self.fileDir = initData.fileDir
      super().__init__(initData, **kwargs)
    self.readFunc = readFunc
    self.allowAbsolute = allowAbsolute

  def __missing__(self, key):
    key = str(key)
    exists = super().get(key, self._UNSET)
    if exists is not self._UNSET:
      return exists
    if self.fileDir is None:
      raise KeyError(f'"{key}" is not in dict and no backing file system was provided')
    pathKey = Path(key)
    isAbsolute = pathKey.is_absolute()
    if not self.allowAbsolute and isAbsolute:
      raise KeyError(f'Directory paths must be relative, received {key}')
    testPath = pathKey if isAbsolute else self.fileDir/key
    candidates = list(testPath.parent.glob(testPath.name))
    if len(candidates) != 1:
      grammar = ': ' if len(candidates) else ''
      raise KeyError(f'"{key}" corresponds to {len(candidates)} files{grammar}{", ".join(c.name for c in candidates)}')
    else:
      file = candidates[0]
      ret = self[key] = self.readFunc(file)
    return ret

  def get(self, key, default=None):
    ret = super().get(key, self._UNSET)
    if ret is self._UNSET:
      # See if directory has this data
      try:
        ret = self[key]
      except KeyError:
        return default
    return ret


def deprecateKwargs(**toDeprecate):
  def deco(func):
    @wraps(func)
    def inner(*args, **kwargs):
      usedDeprecated = set(toDeprecate) & set(kwargs)
      if usedDeprecated:
        grammar = 'is' if len(usedDeprecated) == 1 else 'are'
        replacements = {k: toDeprecate[k] for k in usedDeprecated if toDeprecate[k] is not None}
        msg = f'{", ".join(usedDeprecated)} {grammar} deprecated and will be removed in a future release.'
        if replacements:
          for orig, replace in replacements.items():
            kwargs[replace] = kwargs[orig]
            del kwargs[orig]
          msg += f' Use the following replacement guide: {replacements}'
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
      return func(*args, **kwargs)
    return inner
  return deco



from typing import Sequence, Any
from warnings import warn

import numpy as np
import pandas as pd
from pyqtgraph.Qt import QtWidgets, QtCore, QtGui
from pyqtgraph.parametertree import Parameter

from s3a._io.helpers import serialize, deserialize
from s3a.constants import PRJ_CONSTS, REQD_TBL_FIELDS, PRJ_ENUMS
from s3a.models.tablemodel import ComponentMgr
from s3a.shared import SharedAppSettings
from s3a.structures import TwoDArr

__all__ = ['CompTableView']

from utilitys import ParamEditor, EditorPropsMixin, RunOpts, ParamContainer, DeferredActionStackMixin as DASM, PrjParam
from utilitys.params.pgregistered import PgParamDelegate

Signal = QtCore.Signal

class SerDesDelegate(QtWidgets.QStyledItemDelegate):
  def __init__(self, param: PrjParam, asLineEdit=False, parent=None):
    """
    Creates a string-editable item delegate which uses ComponentIO's serialize and deserialize methods to convert to
    actual values.
    :param param: PrjParam which dictates the serialization/deserialization of this value
    :param asLineEdit: Whether to give this delegate a line editor text edit to set the serialized form of value
    :parent: Parent widget
    """
    super().__init__(parent)
    self.asLineEdit=asLineEdit
    self.param = param

  def createEditor(self, parent, option, index: QtCore.QModelIndex):
    editorType = 'str' if self.asLineEdit else 'text'
    pgParam = Parameter.create(name='dummy', type=editorType, value=index.data(QtCore.Qt.DisplayRole))
    editor = pgParam.itemClass(pgParam, 0).makeWidget()
    editor.setParent(parent)
    editor.setMaximumSize(option.rect.width(), option.rect.height())
    return editor

  def setModelData(self, editor: QtWidgets.QWidget,
                   model: QtCore.QAbstractTableModel,
                   index: QtCore.QModelIndex):
    values, errs = deserialize(self.param, [editor.value()])
    if not len(values):
      raise ValueError(f'Error during deserialize:\n{errs[0]}')
    value = values[0]
    model.setData(index, value, QtCore.Qt.EditRole)

  def setEditorData(self, editor: QtWidgets.QWidget, index):
    value = index.data(QtCore.Qt.EditRole)
    strVals, strErrs = serialize(self.param, [value])
    if len(strErrs):
      raise ValueError(f'Error during serialize:\n{strErrs[0]}')
    strVal = strVals[0]
    editor.setValue(strVal)

  def updateEditorGeometry(self, editor: QtWidgets.QWidget,
                           option: QtWidgets.QStyleOptionViewItem,
                           index: QtCore.QModelIndex):
    editor.setGeometry(option.rect)

class MinimalTableModel(ComponentMgr):
  """
  The CheckState flag is only possible on a minimal popup table, but Qt makes it
  challenging to override just this portion of the ComponentMgr class. So, this simple
  subclass maintains a checkstate and checkable columns on top of the traditional
  component manager.
  """

  def __init__(self):
    super().__init__()
    # Map true/false to check state
    self.csMap = {True: QtCore.Qt.Checked, False: QtCore.Qt.Unchecked}

    # Since the minimal model only contains one row, everything checkable can be on a
    # per-column basis
    self.checkedColIdxs = set()

  def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlags:
    return super().flags(index) | QtCore.Qt.ItemIsUserCheckable

  def data(self, index: QtCore.QModelIndex, role: int) -> Any:
    if role == QtCore.Qt.CheckStateRole:
      return self.csMap[index.column() in self.checkedColIdxs]
    return super().data(index, role)

  def setData(self, index, value, role=QtCore.Qt.EditRole) -> bool:
    if role == QtCore.Qt.CheckStateRole:
      if value == QtCore.Qt.Checked:
        self.checkedColIdxs.add(index.column())
      else:
        self.checkedColIdxs.remove(index.column())
      return True
    yield from super().setData(index, value, role)


class PopupTableDialog(QtWidgets.QDialog):
  def __init__(self, *args):
    super().__init__(*args)
    self.setModal(True)
    # -----------
    # Table View
    # -----------
    self.tbl = CompTableView(minimal=True)
    self.model = MinimalTableModel()
    self.tbl.setModel(self.model)

    # -----------
    # Warning Message
    # -----------
    self.reflectDelegateChange()
    self.warnLbl = QtWidgets.QLabel(
      'Check the boxes for data you wish to propagate to all selected cells.\n'
      'Selections where every row has the same data are checkd by default.', self)
    self.warnLbl.setStyleSheet("font-weight: bold; color:red; font-size:14")

    # -----------
    # Widget buttons
    # -----------
    self.applyBtn = QtWidgets.QPushButton('Apply')
    self.closeBtn = QtWidgets.QPushButton('Close')

    # -----------
    # Widget layout
    # -----------
    btnLayout = QtWidgets.QHBoxLayout()
    btnLayout.addWidget(self.applyBtn)
    btnLayout.addWidget(self.closeBtn)

    centralLayout = QtWidgets.QVBoxLayout()
    centralLayout.addWidget(self.warnLbl)
    centralLayout.addWidget(self.tbl)
    centralLayout.addLayout(btnLayout)
    self.setLayout(centralLayout)
    self.setMinimumWidth(self.tbl.width())

    # -----------
    # UI Element Signals
    # -----------
    self.closeBtn.clicked.connect(self.close)
    self.applyBtn.clicked.connect(self.accept)

  def reflectDelegateChange(self):
    # TODO: Find if there's a better way to see if changes happen in a table
    self.titles = np.array(list([f.name for f in self.tbl.tableData.allFields]))
    self.tbl.setColDelegates()
    # Avoid loop scoping
    def wrapper(col):
      def onUpdate():
        self.model.checkedColIdxs.add(col)
      return onUpdate
    for colIdx in range(len(self.titles)):
      deleg = self.tbl.itemDelegateForColumn(colIdx)
      deleg.commitData.connect(wrapper(colIdx))

  @property
  def data(self):
    return self.tbl.mgr.compDf.iloc[[0],:]

  def setData(self, compDf: pd.DataFrame, colIdxs: Sequence[int],
              dirtyColIdxs: Sequence[int]=None):
    # New selection, so reset dirty columns
    if dirtyColIdxs is None:
      dirtyColIdxs = []
    # Hide columns that weren't selected by the user since these changes
    # Won't propagate
    self.model.checkedColIdxs.clear()
    self.model.checkedColIdxs.update(dirtyColIdxs)

    for ii in range(len(self.titles)):
      if ii not in colIdxs:
        self.tbl.hideColumn(ii)
      else:
        self.tbl.showColumn(ii)
    self.tbl.mgr.rmComps()
    self.tbl.mgr.addComps(compDf, addtype=PRJ_ENUMS.COMP_ADD_AS_MERGE)

  def reject(self):
    # On dialog close be sure to unhide all columns / reset dirty cols
    for ii in range(len(self.titles)):
      self.tbl.showColumn(ii)
    super().reject()

class CompTableView(DASM, EditorPropsMixin, QtWidgets.QTableView):
  __groupingName__ = 'Component Table'
  """
  Table for displaying :class:`ComponentMgr` data.
  """
  sigSelectionChanged = Signal(object)

  def __initEditorParams__(self, shared: SharedAppSettings):
    self.props = ParamContainer()
    self.tableData = shared.tableData
    shared.generalProps.registerProp(PRJ_CONSTS.PROP_SHOW_TBL_ON_COMP_CREATE,
                                                               container=self.props)

    self.toolsEditor = ParamEditor.buildClsToolsEditor(type(self), name='Component Table Tools')

    with shared.generalProps.setBaseRegisterPath(self.__groupingName__):
      proc, params = shared.generalProps.registerFunc(
        self.setVisibleColumns, runOpts=RunOpts.ON_CHANGED, nest=False,
        returnParam=True, visibleColumns=[])
    def onChange(*_args):
      params.child('visibleColumns').setLimits([f.name for f in shared.tableData.allFields])
    onChange()
    shared.tableData.sigCfgUpdated.connect(onChange)

  def __init__(self, *args, minimal=False):
    """
    Creates the table.

    :param minimal: Whether to make a table with minimal features.
       If false, creates extensible table with context menu options.
       Otherwise, only contains minimal features.
    """
    super().__init__(*args)

    self.setStyleSheet("QTableView { selection-color: white; selection-background-color: #0078d7; }")
    self._prevSelRows = np.array([])
    self.setSortingEnabled(True)

    self.mgr = ComponentMgr()
    self.minimal = minimal
    self.setModel(self.mgr)
    self.setColDelegates()

    if not minimal:
      self.popup = PopupTableDialog(*args)
      # Create context menu for changing table rows
      self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
      cursor = QtGui.QCursor()
      self.customContextMenuRequested.connect(lambda: self.menu.exec_(cursor.pos()))

    self.instIdColIdx = self.tableData.allFields.index(REQD_TBL_FIELDS.INST_ID)

  def setVisibleColumns(self, visibleColumns: Sequence[str]):
    """
    Determines which columns to show. All unspecified columns will be hidden.
    :param visibleColumns:
      pType: checklist
    """
    for ii, col in enumerate(self.mgr.colTitles):
      self.setColumnHidden(ii, col not in visibleColumns)

  def setColDelegates(self):
    for ii, field in enumerate(self.tableData.allFields):
      curType = field.pType.lower()
      curval = field.value
      paramDict = dict(type=curType, default=curval, **field.opts)
      if curType == 'enum':
        paramDict['type'] = 'list'
        paramDict.update(values=list(type(curval)))
      elif curType == 'prjparam':
        paramDict['type'] = 'list'
        paramDict.update(values=list(curval.group))
      elif curType == 'bool':
        # TODO: Get checkbox to stay in table after editing for a smoother appearance.
        #   For now, the easiest solution is to just use dropdown
        paramDict['type'] = 'list'
        paramDict.update(values={'True': True, 'False': False})
      try:
        self.setItemDelegateForColumn(ii, PgParamDelegate(paramDict, self))
      except TypeError:
        # Parameter doesn't have a registered pyqtgraph editor, so default to
        # generic serdes editor
        self.setItemDelegateForColumn(ii, SerDesDelegate(field, parent=self))

    self.horizontalHeader().setSectionsMovable(True)

  # When the model is changed, get a reference to the ComponentMgr
  def setModel(self, modelOrProxy: QtCore.QAbstractTableModel):
    super().setModel(modelOrProxy)
    try:
      # If successful we were given a proxy model
      self.mgr = modelOrProxy.sourceModel()
    except AttributeError:
      self.mgr = modelOrProxy

  def selectionChanged(self, curSel: QtCore.QItemSelection, prevSel: QtCore.QItemSelection):
    """
    When the selected rows in the table change, this retrieves the corresponding previously
    and newly selected IDs. They are then emitted in sigSelectionChanged.
    """
    super().selectionChanged(curSel, prevSel)
    selectedIds = []
    selection = self.selectionModel().selectedIndexes()
    for item in selection:
      selectedIds.append(item.sibling(item.row(),self.instIdColIdx).data(QtCore.Qt.EditRole))
    newRows = pd.unique(selectedIds)
    if np.array_equal(newRows, self._prevSelRows):
      return
    self._prevSelRows = newRows
    self.sigSelectionChanged.emit(newRows)

  def removeSelectedRows_gui(self):
    if self.minimal: return

    idList = [idx.siblingAtColumn(self.instIdColIdx).data(QtCore.Qt.EditRole)
              for idx in self.selectedIndexes()]
    if len(idList) == 0:
      return
    # Make sure the user actually wants this
    idList = pd.unique(idList)
    dlg = QtWidgets.QMessageBox()
    confirm  = dlg.question(self, 'Remove Rows', f'Are you sure you want to remove {len(idList)} selected row(s)?',
                 dlg.Yes | dlg.Cancel, dlg.Yes)
    if confirm == dlg.Yes:
      # Proceed with operation
      # Since each selection represents a row, remove duplicate row indices
      self.mgr.rmComps(idList)
      self.clearSelection()

  def ids_rows_colsFromSelection(self, excludeNoEditCols=True, warnNoneSelection=True,
                                 selectedIdxs: Sequence[QtCore.QModelIndex]=None):
    """
    Returns Nx3 np array of (ids, rows, cols) from current table selection. Ids refer to
    the 'Instance ID' column, while rows refer to the placement in the visible table.
    Thus, sorting the table will not affect `id` values, but will affect `row` values.

    :param excludeNoEditCols: Whether to consider columns that do not allow editing, like
      Instance ID and Vertices
    :param warnNoneSelection: Whether to raise a warning if no values are selected
    :param selectedIdxs: Selection to get ids, rows, and cols from. If *None*, defaults to
      the current table selection (self.selectedIndexes())
    """
    if selectedIdxs is None:
      selectedIdxs = self.selectedIndexes()
    retLists = [] # (Ids, rows, cols)
    for idx in selectedIdxs:
      row = idx.row()
      # 0th row contains instance ID
      # TODO: If the user is allowed to reorder columns this needs to be revisited
      idAtIdx = idx.siblingAtColumn(self.instIdColIdx).data(QtCore.Qt.EditRole)
      retLists.append([idAtIdx, row, idx.column()])
    retLists = np.array(retLists)
    if excludeNoEditCols and len(retLists) > 0:
      # Set diff will eliminate any repeats, so use a slower op that at least preserves
      # duplicates
      retLists = retLists[~np.isin(retLists[:,2], self.mgr.noEditColIdxs)]
    if len(retLists) == 0:
      retLists.shape = (-1,3)
    if warnNoneSelection and len(retLists) == 0:
      warn('No editable columns selected.', UserWarning)
    return retLists

  def setSelectedCellsAsFirst(self):
    """
    Sets all cells in the selection to be the same as the first row in the selection.
    See the project wiki for a detailed description
    """
    selection = self.ids_rows_colsFromSelection()
    overwriteData = self.mgr.compDf.loc[selection[0,0]]
    self.setSelectedCellsAs(selection, overwriteData)

  def setSelectedCellsAs_gui(self, selectionIdxs: TwoDArr=None):
    """
    Sets all cells in the selection to the values specified in the popup table. See
    the project wiki for a detailed description
    """
    if selectionIdxs is None:
      selectionIdxs = self.ids_rows_colsFromSelection()
    overwriteData = self.mgr.compDf.loc[[selectionIdxs[0,0]]].copy()
    with self.actionStack.ignoreActions():
      self.popup.setData(overwriteData, pd.unique(selectionIdxs[:,2]),
                         self._getDupDataCols(selectionIdxs))
      wasAccepted = self.popup.exec_()
    # Convert to list or isin() check below fails
    dirtyCols = list(self.popup.model.checkedColIdxs)
    if not wasAccepted or not dirtyCols:
      return

    selectionIdxs = selectionIdxs[np.isin(selectionIdxs[:,2], dirtyCols)]
    self.setSelectedCellsAs(selectionIdxs, self.popup.data)

  def _getDupDataCols(self, selectionIdxs: np.ndarray):
    """
    From a selection of components, returns the column names which have the same data in
    every row. This is useful for determining which columns should be propagated by
    default when performing a "set cells" operation

    :param selectionIdxs: Selection list of dataframe cells to consider. See
      `ids_rows_colsFormSelection` for a description of this array
    :return: List of column idxs which have only duplicate data selected, i.e. all
    entries in every row of that column are the same
    """
    dupCols = []
    # First, find out all rows for each column to select, then test data at those rows
    for col in np.unique(selectionIdxs[:, 2]):
      name = self.mgr.compDf.columns[col]
      ids = selectionIdxs[selectionIdxs[:, 2] == col, 0]
      data = self.mgr.compDf.loc[ids, name]
      try:
        numDups = len(pd.unique(data))
      except TypeError:
        # Occurs for unhashable types
        numDups = 2
      if numDups <= 1:
        dupCols.append(col)
    return dupCols

  def setSelectedCellsAs(self, selectionIdxs: TwoDArr, overwriteData: pd.DataFrame):
    """
    Overwrites the data from rows and cols with the information in *overwriteData*.
    Each (id, row, col) index is treated as a single index
    :param selectionIdxs: Selection idxs to overwrite. If *None*, defaults to
      current selection.
    :param overwriteData: What to fill in the overwrite locations. If *None*, a popup table
      is displayed and its data is used.
    """
    if self.minimal: return

    if len(selectionIdxs) == 0:
      return
    overwriteData = overwriteData.squeeze()
    uniqueIds = pd.unique(selectionIdxs[:,0])
    newDataDf = self.mgr.compDf.loc[uniqueIds].copy()
    # New data ilocs will no longer match, fix this using loc + indexed columns
    colsForLoc = self.mgr.compDf.columns[selectionIdxs[:,2]]
    for idxTriplet, colForLoc in zip(selectionIdxs, colsForLoc):
      newDataDf.at[idxTriplet[0], colForLoc] = overwriteData.iat[idxTriplet[2]]
    self.mgr.addComps(newDataDf, addtype=PRJ_ENUMS.COMP_ADD_AS_MERGE)


import typing as t

import cv2 as cv
import numpy as np
import pandas as pd
from scipy.ndimage import maximum_filter

from s3a.constants import PRJ_ENUMS, REQD_TBL_FIELDS as RTF
from s3a.generalutils import getCroppedImg, bboxIou
from s3a.structures import ComplexXYVertices
from utilitys import ProcessIO, fns, AtomicProcess


def get_component_images(image: np.ndarray, components: pd.DataFrame):
  """
  From a main image and dataframe of components, adds an 'img' column to `components` which holds the
  subregion within the image each component occupies.
  """
  imgs = [getCroppedImg(image, verts.stack(), 0) for verts in components[RTF.VERTICES]]
  return ProcessIO(subimages=imgs)

def _focusedResultConverter(result, component: pd.Series):
  out = result
  if not isinstance(result, ProcessIO):
    out = ProcessIO(image=result)
  if 'components' not in out:
    mask = out['image']
    newComp = component.copy()
    if mask is not None:
      offset = newComp[RTF.VERTICES].stack().min(0)
      newVerts = ComplexXYVertices.fromBwMask(mask)
      for v in newVerts: v += offset
      newComp[RTF.VERTICES] = newVerts
    out['components'] = fns.serAsFrame(newComp)
  out['addType'] = PRJ_ENUMS.COMP_ADD_AS_MERGE
  return out

def _dispatchFactory(func, resultConverter: t.Callable[[t.Union[dict, t.Any], pd.Series], ProcessIO]=None):
  def dispatcher(image: np.ndarray, components: pd.DataFrame, **kwargs):
    compList = []
    kwargs.update(image=image)
    result = ProcessIO()
    for ii, comp in components.iterrows():
      kwargs.update(component=comp)
      # TODO: Determine appropriate behavior. For now, just remember last result metadata other than comps
      result = func(**kwargs)
      if resultConverter is not None:
        result = resultConverter(result, comp)
      compList.append(result.pop('components'))
    if compList:
      # Concat fails with empty list
      outComps = pd.concat(compList, ignore_index=True)
    else:
      outComps = components.drop(components.index).copy()
    out = ProcessIO(**result, components=outComps)
    return out

  proc = AtomicProcess(dispatcher, docFunc=func, ignoreKeys=['component'])
  proc.input['components'] = ProcessIO.FROM_PREV_IO
  return proc

def pts_to_components(matchPts: np.ndarray, component: pd.Series):
  numOutComps = len(matchPts)
  if numOutComps == 0:
    ret = fns.serAsFrame(component).copy()
    return ret.loc[[]]
  # Explicit copy otherwise all rows point to the same component
  outComps = pd.concat([fns.serAsFrame(component)]*numOutComps, ignore_index=True).copy()
  origOffset = component[RTF.VERTICES].stack().min(0)
  allNewverts = []
  for ii, pt in zip(outComps.index, matchPts):
    newVerts = [verts-origOffset+pt for verts in outComps.at[ii, RTF.VERTICES]]
    allNewverts.append(ComplexXYVertices(newVerts))
  outComps[RTF.VERTICES] = allNewverts
  return outComps


@fns.dynamicDocstring(metricTypes=[d for d in dir(cv) if d.startswith('TM')])
def _cv_template_match(component: pd.Series, image: np.ndarray, viewbox: np.ndarray,
                       threshold=0.8, metric='TM_CCOEFF_NORMED', area='viewbox',
                       refOverlapThresh=0.5):
  """
  Performs template matching using default opencv functions
  :param component: Template component
  :param image: Main image
  :param threshold:
    helpText: Cutoff point to consider a matched template
    limits: [0, 1]
    step: 0.1
  :param metric:
    helpText: Template maching metric
    pType: list
    limits: {metricTypes}
  :param area:
    helpText: Where to apply the new components
    pType: list
    limits:
      - image
      - viewbox
  :param refOverlapThresh:
    helpText: "How much overlap can exist between the found bounding box and reference box. This helps
    prevent detecting the original as a new component"
    limits: [0,1]
    step: 0.1
  """
  template, templateBbox = getCroppedImg(image, component[RTF.VERTICES].stack())
  if area == 'viewbox':
    image, coords = getCroppedImg(image, viewbox)
  else:
    coords = np.array([[0,0], image.shape[:2][::-1]])
  grayImg = image if image.ndim < 3 else cv.cvtColor(image, cv.COLOR_RGB2GRAY)
  if template.ndim > 2:
    template = cv.cvtColor(template, cv.COLOR_RGB2GRAY)
  templateShp = template.shape[:2]
  if np.any(np.array(templateShp) > np.array(grayImg.shape)):
    raise ValueError('Search area cannot be smaller than template size.\n'
                     f'Search area size: {grayImg.shape}, template size: {templateShp}')

  metric = getattr(cv, metric)
  res = cv.matchTemplate(grayImg, template, metric)
  maxFilter = maximum_filter(res, templateShp)
  # Non-max suppression to remove close-together peaks
  res[maxFilter > res] = 0
  loc = np.nonzero(res >= threshold)
  scores = res[loc]
  matchPts = np.c_[loc[::-1]] + coords[[0]]
  # Don't allow matches on top of originals
  ious = []
  for pt in matchPts:
    ious.append(bboxIou(templateBbox, np.vstack([pt, pt + templateShp[::-1]])))
  keep = np.array(ious) < refOverlapThresh
  scores = scores[keep]
  matchPts = matchPts[keep]
  return ProcessIO(scores=scores, matchImg=maxFilter, components=pts_to_components(matchPts, component))

def cv_template_match_factory():
  return _dispatchFactory(_cv_template_match)